#include <windows.h>
#include <iostream.h>
#include <stdlib.h>
#include <time.h>
#include <vector>

#include "usb100.h"     // DDK
#include <setupapi.h>   // DDK
#include <winioctl.h>


#include <logfile.h>

//LogFile Log("Riositude.log");

#include "rio500.h"
#include "../RioUtil.h"


static GUID GUID_CLASS_RIO500 = { 0xa5dcbf10, 0x6530, 0x11d2, 0x90, 0x1f, 0x00, 0xc0, 0x4f, 0xb9, 0x51, 0xed };
          //{0x23712389, 0x6d42, 0x4067, 0x8d, 0xe6, 0x8b, 0x74, 0xb6, 0x36, 0x85, 0x27};

// OK
Rio500::Rio500()
{
  //---Log << "Initializing Rio500 Object!\n";
  rio500 = rio500Read = rio500Write = INVALID_HANDLE_VALUE;
  strcpy(fontName, "MS Sans Serif");
  fontSize = 8;

  lastError = S_OK;
}

// OK
Rio500::~Rio500()
{
  disconnect();
}

// OK
bool Rio500::connect()
{
  //---Log << "Connecting!\n";
  lastError = S_OK;
  try
  {
    //---Log << "  Opening driver.\n";
    int i = openDriver();
    if(i != 0)
    {
      //---Log << "  Trying again.";
      i = openDriverRaw();
      if(i != 0)
      {
        Sleep(50);
        i = openDriver();
        if(i != 0)
        {
          i = openDriverRaw();
        }
      }
    }
  
    if(i != 0)
      throw RIO_E_COULDNT_OPEN_DEVICE;
  }
  catch(HRESULT e)
  {
    //---Log << "  Error: " << e << "\n";
    lastError = e;
    return false;
  }
  return true;
}

// OK
bool Rio500::disconnect()
{
  //---Log << "Disconnecting!\n";
  lastError = S_OK;
  if(rio500 == INVALID_HANDLE_VALUE) return true;
  try
  {
    endComm();
    closeDriver();
  }
  catch(HRESULT e)
  {
    //---Log << "  Error: " << e << "\n";
    lastError = e;
    return false;
  }
  return true;
}

// OK
bool Rio500::hasExternalCard()
{
  //---Log << "Checking for external card!\n";
  lastError = S_OK;
  HRESULT hr = S_OK;

  bool retVal;

  try
  {
    // Make sure the driver is opened
	  if (!IsDriverOpened())
      throw RIO_E_DRIVER_NOT_OPEN;
      //---Log << "  Driver is open!\n";

    // detect flash card
    int val = (long) ((sendCommand(REQ_NOP, 0x0000, 0x0000) & 0x40000000) >> 30);
    retVal = (val ? true : false);

    // Wait a bit...
    NOP();
    NOP();
  }
  catch (HRESULT e)
  {
    //---Log << "  Error: " << e << "\n";
    lastError = e;
    return false;
  }

  //---Log << "  Retval: " << retVal << "\n";
  return retVal;
}

// OK
int Rio500::getAvailableMemory(int card)
{
  //---Log << "Retrieving available memory on " << card << "!\n";
  lastError = S_OK;
  int val;
  try
  {
    if(!IsDriverOpened())
      throw RIO_E_DRIVER_NOT_OPEN;
  
    val = sendCommand(REQ_QUERY_FREE_MEM, 0x0000, card);
    NOP();
    NOP();
  }
  catch(HRESULT e)
  {
    //---Log << "  Error: " << e << "\n";
    lastError = e;
    return -1;
  }
  //---Log <<"  Val: " << val << "\n";
  return val;
}

// OK
int Rio500::getTotalMemory(int card)
{
  //---Log << "Retrieving total memory on " << card << "!\n";
  lastError = S_OK;
  int val;
  try
  {
    // Make sure the driver is open
    if (!IsDriverOpened())
      throw RIO_E_DRIVER_NOT_OPEN;

    // New mem-status block
    //---Log << "  Memsetting...\n";
    RioMemStatus memStatus;
    memset(&memStatus, 0, sizeof(RioMemStatus));

    // Get the mem block
    sendCommand(REQ_SET_TO_MEM_LOCATION, 1,card);

    //---Log << "  IORequest!\n";
    IORequest(DIR_IN, REQ_GET_MEM_STATUS, 0, 0, sizeof(RioMemStatus), (BYTE*)&memStatus);

    //---Log << "  Done!\n";
    // Wait.  Don't know if we need to, but every other function in this object does.
    NOP();
    NOP();

    // Multiply to get the result.
    val = memStatus.wNumBlocks * memStatus.wBlockSize;
  }
  catch(HRESULT e)
  {
    //---Log << "  Error: " << e << "\n";
    lastError = e;
    return -1;
  }

  //---Log << "  Val: " << val << "\n";
  return val;
}

// OK
int Rio500::getFirmwareVersion()
{
  //---Log << "Get Firmware Version" << "\n";
  lastError = S_OK;
  int val;
  try
  {
    if(!IsDriverOpened())
      throw RIO_E_DRIVER_NOT_OPEN;

    val = (sendCommand(REQ_FIRMWARE, 0x0000, 0x0000) & 0xFFFF);
    NOP();
    NOP();
  }
  catch(HRESULT e)
  {
    lastError = e;
    return -1;
  }
  //---Log << "  Val: " << val << "\n";
  return val;
}

// OK
bool Rio500::format(int card)
{
  //---Log << "  Formatting card " << card << "\n";
  try
  {
    // Make sure the driver is opened
	  if (!IsDriverOpened())
      throw RIO_E_DRIVER_NOT_OPEN;

      // Send the format command
		if (sendCommand(REQ_RIO_FORMAT_DEVICE, 0x2185, card) != RSTATUS_OK)
			throw RIO_E_FORMAT_FAILED;

  	NOP();
    NOP();
  }
  catch (HRESULT e)
  {
    //---Log << "  Error: " << e << "\n";
    lastError = e;
    return false;
  }
  return true;
}

// OK
bool Rio500::addFolder(char *name, int card)
{
  lastError = S_OK;
  try
  {
      // Make sure the driver is opened
    if (!IsDriverOpened())
      throw RIO_E_DRIVER_NOT_OPEN;

    // Build a folder object
    RioFolderEntry folder;
    ZeroMemory(&folder, sizeof(RioFolderEntry));

    // Put in the name..
    strcpy(folder.szName1, name);

    // Build a bitmap for this folder name in a variant
    if (!buildTextBitmap(name, folder.Bitmap.Image, folder.Bitmap.wNumBlocks))
      throw RIO_E_CREATE_BITMAP_FAILED;

    folder.dwDunno3 = 0x002100FF;
    folder.dwTime = time(NULL);

    strncat(folder.szName2, folder.szName1, 127);


    RioFolderList folders;

    // If this is not the first folder, read the existing folder entries
    if (!isFirstFolder(card))
        folders = readFolderEntries(card);

    // Get number of folders        
    DWORD dwNumFolders = folders.size();

    // Make sure there are less than 8 folders already present
    if (dwNumFolders > 7)
      throw RIO_E_MAX_FOLDERS;

    // Wait a bit..    
    NOP();
    NOP();

    // Allocate an empty folder block
    if (!allocateEmptyFolderBlock(dwNumFolders, card))
      throw RIO_E_FOLDER_ALLOC_FAILED;

    // Wait a bit..
    NOP();
    NOP();

    // Find out where the new empty folder block is 
    DWORD dwSendBlockLoc = sendCommand(REQ_QUERY_OFFSET_LAST_WRITE, 0x0000, 0x0000);  // no lFlash necessary ?

    // Put empty folder block location in current folder        
    folder.wOffset = dwSendBlockLoc;

    // Add the folder entry to the list of folders
    folders.push_back(folder);

    // Write all the folder entries back to the Rio
    if (!writeFolderEntries(folders, card))
        throw RIO_E_FOLDER_WRITE_FAILED;

    // Wait a bit
    NOP();
    NOP();

    // Find out where the folder block was written to
    DWORD dwFolderBlockLoc = sendCommand(REQ_QUERY_OFFSET_LAST_WRITE, 0x0000, 0x0000);  // no lFlash necessary ?

    // Tell the Rio where to find the folder block        
    if (!sendFolderLocation(dwFolderBlockLoc, dwNumFolders, card))
      throw RIO_E_COULDNT_SEND_FOLDER_LOC;

    // End folder upload
    sendCommand(REQ_END_FOLDER_TRANSFERS, 0x0000, card);

  }
  catch (HRESULT e)
  {
    lastError = e;
    return false;
  }
  return true;
}

// OK
bool Rio500::renameFolder(int folderNum, char *newName, int card)
{
  lastError = S_OK;
  try
  {
      // Make sure the driver is opened
	  if (!IsDriverOpened())
      throw RIO_E_DRIVER_NOT_OPEN;

    // Read all the folders
    RioFolderList folders = readFolderEntries(card);

    // Make sure we have a valid folder number
    if (folderNum >= folders.size())
        throw RIO_E_INVALID_FOLDER;

    // Build the bitmap for the folder name
    if(!buildTextBitmap(newName, folders[folderNum].Bitmap.Image, folders[folderNum].Bitmap.wNumBlocks))
      throw RIO_E_CREATE_BITMAP_FAILED;

    // Convert the BSTR name to an ASCII name...
    memset(folders[folderNum].szName1, 0, 362); // UPDATE BY JJ: Prevents corrpution
    strcpy(folders[folderNum].szName1, newName);

    // ... and copy it to the secondary name location
    *folders[folderNum].szName2 = NULL;
    strncat(folders[folderNum].szName2, folders[folderNum].szName1, 127);

    // Write the folder entries back to the Rio
    writeFolderEntries(folders, card);

    // Wait a bit...
    NOP();
    NOP();

    // Get the new folder entries location
    DWORD   dwFolderBlockLoc = sendCommand(REQ_QUERY_OFFSET_LAST_WRITE, 0x0000, 0x0000);  // no lFlash necessary ?

    // Tell the Rio where they're at
    sendFolderLocation(dwFolderBlockLoc, folderNum, card);

    // End folder transfer
    sendCommand(REQ_END_FOLDER_TRANSFERS, 0x0000, card);
  }
  catch (HRESULT e)
  {
    lastError = e;
    return false;
  }
  return true;
}

// OK
bool Rio500::deleteFolder(int folderNum, int card)
{
  try
  {
    // Make sure the driver is opened
	  if (!IsDriverOpened())
      throw RIO_E_DRIVER_NOT_OPEN;

    // Read the existing folder list
    RioFolderList   folders = readFolderEntries(card);
    int i;

    // Make sure we have a valid folder number
    if (folderNum >= folders.size())
      throw RIO_E_INVALID_FOLDER;

    // Read all songs in this folder...
    RioSongList   songs = readSongEntries(folders[folderNum], folderNum, card);

    // ... and delete them.  I don't know why this works, but it does.
    for (i = 0; i < songs.size(); i++)
      sendCommand(REQ_SET_WRITE_ADDRESS, ((folderNum << 8) | i), card);

    // Delete the folder
    sendCommand(REQ_SET_WRITE_ADDRESS, (folderNum << 8), card);

    // Remove the folder entry from the folder list
    folders.erase(folders.begin() + folderNum);

    // Write the folder entries back to the Rio
    writeFolderEntries(folders,card);

    // Wait a bit..
    NOP();
    NOP();

    // Get the new folder entries location
    DWORD   dwFolderBlockLoc = sendCommand(REQ_QUERY_OFFSET_LAST_WRITE, 0x0000, 0x0000);  // no lFlash necessary ?

    // Tell the Rio where they're at
    sendFolderLocation(dwFolderBlockLoc, folderNum, card);

    // End folder transfer
    sendCommand(REQ_END_FOLDER_TRANSFERS, 0x0000, card);
  }
  catch (HRESULT e)
  {
    lastError = e;
    return false;
  }

  return true;
}

// OK
RioFolderList Rio500::getFolders(int card)
{
  //---Log << "Getting Folders on " << card << "\n";
  lastError = S_OK;
  RioFolderList list;
  try
  {
    if(!IsDriverOpened()) throw RIO_E_DRIVER_NOT_OPEN;

    list = readFolderEntries(card);
  }
  catch(HRESULT e)
  {
    //---Log << "  Error: " << e << "\n";
    lastError = e;
    return list;
  }
  //---Log << "Success!\n";
  return list;
}

bool Rio500::putFolders(RioFolderList list, int card)
{
  lastError = S_OK;
  try
  {
      // Make sure the driver is opened
    if (!IsDriverOpened())
      throw RIO_E_DRIVER_NOT_OPEN;

    // Write all the folder entries back to the Rio
    if (!writeFolderEntries(list, card))
        throw RIO_E_FOLDER_WRITE_FAILED;

    // Wait a bit
    NOP();
    NOP();

    // Find out where the folder block was written to
    DWORD dwFolderBlockLoc = sendCommand(REQ_QUERY_OFFSET_LAST_WRITE, 0x0000, 0x0000);  // no lFlash necessary ?

    // Tell the Rio where to find the folder block        
    if (!sendFolderLocation(dwFolderBlockLoc, list.size(), card))
      throw RIO_E_COULDNT_SEND_FOLDER_LOC;

    // End folder upload
    sendCommand(REQ_END_FOLDER_TRANSFERS, 0x0000, card);

  }
  catch (HRESULT e)
  {
    lastError = e;
    return false;
  }
  return true;

}

// OK
bool Rio500::addSong(int folderNum, char *name, char *filename, int card, ProgressCallback callback,void *parm)
{
  HRESULT     hr = S_OK;
  HANDLE      hFile = INVALID_HANDLE_VALUE;
  BYTE*       pBlock = NULL;

  bool        retVal = true;

  try
  {
    // Make sure the driver is opened
	  if (!IsDriverOpened())
      throw RIO_E_DRIVER_NOT_OPEN;

      // Build a new song object
    RioSongEntry song;
    ZeroMemory(&song, sizeof(RioSongEntry));

      // Put the name, song file name and folder number in the object
    strcpy(song.szName1, name);
    strcpy(song.szName2, filename);

    //song.put_Folder(lFolder);

    // Build a bitmap for the song name
    if (!buildTextBitmap(name, song.Bitmap.Image, song.Bitmap.wNumBlocks))
      throw RIO_E_CREATE_BITMAP_FAILED;

    // Try to open the file
    hFile = CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);

    // Make sure it opened successfully
    if (hFile == INVALID_HANDLE_VALUE)
      throw RIO_E_FILE_NOT_FOUND;

    // Build a new song block
    pBlock = new BYTE[RIO_HUGEBLOCKSIZE];

    DWORD dwFileSize;

    // Get the file size of the song
    dwFileSize = GetFileSize(hFile, NULL);
    //---Log << "Adding song...file size: " << (int)dwFileSize << "\n";

    DWORD dwMemLeft = 0;

    // Get free memory on the Rio.
    dwMemLeft = (long)sendCommand(REQ_QUERY_FREE_MEM, 0x0000, card);
    //---Log << "Free Memory: " << (int)dwMemLeft << "\n";

    // Wait a bit..
    NOP();
    NOP();

    // Make sure there's enough memory for the specified song
    if (dwMemLeft < dwFileSize)
      throw RIO_E_NOT_ENOUGH_MEMORY;
  
    // Read the folder list from the Rio
    RioFolderList folders = readFolderEntries(card);

    // Make sure we have a valid folder number        
    if (folderNum > folders.size()-1)
      throw RIO_E_INVALID_FOLDER;

    //---Log << "Folder Num: " << folderNum << "\n";
  
    // Calculate number of data blocks and remainder size
    /*DWORD dwNumBlocks  = dwFileSize / RIO_HUGEBLOCKSIZE;
    DWORD dwAddBlocks  = (dwFileSize % RIO_HUGEBLOCKSIZE) / RIO_BIGBLOCKSIZE;
    DWORD dwRemainder  = (dwFileSize % RIO_HUGEBLOCKSIZE) % RIO_BIGBLOCKSIZE;
    DWORD	dwHowMuchReq = 0;
    DWORD dwHowMuch    = 0;
    DWORD dwNumRead    = 0;
    DWORD	dwWritten    = 0;

    //---Log << "NumBlocks: " << (int)dwNumBlocks << "\n";
    //---Log << "AddBlocks: " << (int)dwAddBlocks << "\n";
    //---Log << "Remainder: " << (int)dwRemainder << "\n";

    // Who knows... 
    sendCommand(REQ_MYSTERY, 0xFFFF, card);


    // Write out all the blocks...
    //---Log << "Entering for loop!\n";
    for (DWORD b = 0; b <= dwNumBlocks; b++)
    {

      // We want to write 1MB blocks or some c*64kB block if less than 1MB remains
      dwHowMuchReq = (b < dwNumBlocks ? RIO_HUGEBLOCKSIZE : dwAddBlocks*RIO_BIGBLOCKSIZE);
      //---Log << "  ------------" << (int)b << "-----------\n";
      //---Log << "  HowMuchReq: " << (int)dwHowMuchReq << "\n";
      if (dwHowMuchReq)
      {
        //---Log << "  Entering while loop!\n";
        while(dwHowMuchReq)
        {
          dwHowMuch = sendCommand(REQ_WRITE_TO_USB, dwHowMuchReq >> 16, dwHowMuchReq & 0xFFFF); // no flash
          dwHowMuchReq -= dwHowMuch;
          //---Log << "    dwHowMuch: " << (int)dwHowMuch << " - Remaining: " << (int)dwHowMuchReq << "\n";

          // Make sure the Rio agrees..
          if (dwHowMuch == 0 || (dwHowMuch & 0x80000000)) // Count "negative" as bad for the hell of it
          {
            //---Log << "    COULDN'T ACCEPT DATA!!!!!!!!\n";
            throw RIO_E_COULDNT_ACCEPT_DATA;
          }
          // Read the data from the file        
          ReadFile(hFile, pBlock, dwHowMuch, &dwNumRead, NULL);

          // Make sure we got what we expect
          if (dwNumRead != dwHowMuch)
            throw RIO_E_SHORT_READ;
          //---Log << "    No Short Read!\n";

          // Write it out to the Rio
          dwWritten = 0;

          //---Log << "    Writing....\n";
          for(int i = 0; i < (dwHowMuch / RIO_BIGBLOCKSIZE); i++)
          {
            dwWritten += bulkWrite(pBlock+i*RIO_BIGBLOCKSIZE, RIO_BIGBLOCKSIZE);
          }

          if(dwHowMuch % RIO_BIGBLOCKSIZE) // Remainder?  How can that be?  Handle it anyway
          {
            //---Log << "    Writing remainder...\n";
            dwWritten += bulkWrite(pBlock+(dwHowMuch-(dwHowMuch%RIO_BIGBLOCKSIZE)), dwHowMuch%RIO_BIGBLOCKSIZE);
          }

          if (dwWritten != dwHowMuch)
	          throw RIO_E_WRITE_FAILED;
          //---Log << "    Write okay!\n";

          // Wait a bit..
          NOP();
          NOP();

          // Fire an update event every 1MB
          if (dwNumBlocks) 
          {
            if(!callback((b*100)/dwNumBlocks, parm))
              throw RIO_E_USER_ABORTED;
          }
        }
        //---Log << "  End while loop!\n";
      }
    }
    //---Log << "End For Loop!\n";

    // If there's a remainder, send it..
    if (dwRemainder)
    {
      //---Log << "We have remainder!\n";
      // Tell the Rio we have a partial block for it..
      dwHowMuchReq = dwRemainder;
      //---Log << "While Loop 2!\n";
      while(dwHowMuchReq)
      {
        dwHowMuch = sendCommand(REQ_WRITE_TO_USB, 0x0000, dwHowMuchReq);
        dwHowMuchReq -= dwHowMuch;
        //---Log << "  dwHowMuch: " << (int)dwHowMuch << " - Remaining: " << (int)dwHowMuchReq << "\n";

        // Make sure it agrees
        if (dwHowMuch == 0 || (dwHowMuch & 0x80000000)) // Count "negative" as bad for the hell of it
        {
          //---Log << "  Coudln't accept data!!!!!!!!\n";
          throw RIO_E_COULDNT_ACCEPT_DATA;
        }

        // Read the data from the file        
        ReadFile(hFile, pBlock, dwHowMuch, &dwNumRead, NULL);

        // Make sure we got what we expect
        if (dwNumRead != dwHowMuch)
          throw RIO_E_SHORT_READ;

        // Write it to the Rio
        if (bulkWrite(pBlock, dwHowMuch) != dwHowMuch)
          throw RIO_E_WRITE_FAILED;
        // Wait a bit..
        NOP();
        NOP();
      }
      //---Log << "End While Loop!\n";
    }*/

    DWORD bytesToWrite = dwFileSize;
    DWORD thisWrite;
    DWORD numRead;
    DWORD written;
    DWORD howMuch;
    int writeFailCount;
    callback(0, parm);

    sendCommand(REQ_MYSTERY, 0xFFFF, card);
    //---Log << "Starting: " << dwFileSize << "\n";
    while(bytesToWrite)
    {
      //---Log << "  Bytes to write: " << bytesToWrite << "\n";
      thisWrite = (bytesToWrite >= RIO_HUGEBLOCKSIZE) ? RIO_HUGEBLOCKSIZE : bytesToWrite;
      //---Log << "    This Write: " << thisWrite << "\n";
      bytesToWrite -= thisWrite;

      //---Log << "    Sending command: ";
      howMuch = sendCommand(REQ_WRITE_TO_USB, thisWrite >> 16, thisWrite & 0xFFFF); // no flash
      if(howMuch == 0)
      {
        howMuch = sendCommand(REQ_WRITE_TO_USB, thisWrite >> 16, thisWrite & 0xFFFF); // no flash
      }

      if(howMuch == 0)
        throw RIO_E_COULDNT_ACCEPT_DATA;
      //---Log << howMuch << "\n";
      ReadFile(hFile, pBlock, thisWrite, &numRead, NULL);

      if(numRead != thisWrite)
        throw RIO_E_SHORT_READ;

      //---Log << "    Writing file\n";
      DWORD sBytesToWrite = thisWrite;
      written = 0;
      BYTE *p=pBlock;
      while(sBytesToWrite)
      {
        thisWrite = (sBytesToWrite >= 0x1000) ? 0x1000 : sBytesToWrite;
        //---Log << "      thisWrite: " << thisWrite << "\n";
        sBytesToWrite -= thisWrite;

        if(thisWrite)
        {
          writeFailCount = 0;
          written = 0;
          while(thisWrite)
          {
            WriteFile(rio500Write, p, thisWrite, &written, 0);
            thisWrite -= written;
            if(thisWrite) writeFailCount++;

            // We'll try FIVE writes.  If they all fail, then we just give up.
            if(writeFailCount == 5)
              throw RIO_E_COULDNT_ACCEPT_DATA;
          }
          p += written;
        }
        //---Log << "      written: " << written << "\n";
      }
      //---Log << written << "\n    Waiting\n";
      NOP();
      NOP();
      //---Log << "    Callback!\n";
      if(!callback(((dwFileSize-bytesToWrite)*100)/dwFileSize, parm))
        throw RIO_E_USER_ABORTED;

    }

    //---Log << "End Loop!\n";

    // Fire an final update event
    callback(100, parm);

    // Get the location of the song
    DWORD   dwSongLoc = sendCommand(REQ_QUERY_OFFSET_LAST_WRITE, 0x0000, 0x0000);  // no lFlash necessary ??

    // Prepare a new song entry

    // reset to begining of the file
    SetFilePointer(hFile,0,NULL,FILE_BEGIN); 

    // Fill in the data for the song
    song.wOffset = dwSongLoc;
    song.dwLength = dwFileSize;
    song.wDunno3 = 0x0020;
    song.dwMP3Sig = (DWORD)getSig(hFile);
    song.dwTime = time(NULL);


    // Read the song list for this folder
    RioSongList songs = readSongEntries(folders[folderNum], folderNum, card);

    // Add this song entry to the list
    songs.push_back(song);

    // Write the song entries back to the Rio
    writeSongEntries(folderNum, songs, card);

    // Wait a bit..
    NOP();
    NOP();

    // Get location of song entries    
    DWORD   dwSongBlockOfs = sendCommand(REQ_QUERY_OFFSET_LAST_WRITE, 0x0000, 0x0000); // no lFlash necessary ?

    // And store them in the folder..
    folders[folderNum].wOffset = dwSongBlockOfs;
    folders[folderNum].wFirstFreeEntryOfs += sizeof(RioFolderEntry);

    // Write out folder entries        
    writeFolderEntries(folders, card);

    // Wait a bit..
    NOP();
    NOP();

    // Get folder location
    DWORD   dwFolderBlockLoc = sendCommand(REQ_QUERY_OFFSET_LAST_WRITE, 0x0000, 0x0000); // no lFlash necessary ?

    // Tell Rio where the folders are
    sendFolderLocation(dwFolderBlockLoc, folderNum,card);

    // End folder transfer
    sendCommand(REQ_END_FOLDER_TRANSFERS, 0x0000, card);
  }
  catch (HRESULT e)
  {
    lastError = e;
    retVal = false;
  }

  // Free song block data
  if (pBlock)
    delete pBlock;

  // Close the song file
  if (hFile != INVALID_HANDLE_VALUE)
    CloseHandle(hFile);

  return retVal;
}

// OK
bool Rio500::renameSong(int folder, int song, char *newname, int card)
{
  try
  {
      // Make sure the driver is opened..
	  if (!IsDriverOpened())
      throw RIO_E_DRIVER_NOT_OPEN;

    // Read the folder entries list
    RioFolderList folders = readFolderEntries(card);

    // Make sure we have a valid folder number
    if (folder >= folders.size())
        throw RIO_E_INVALID_FOLDER;

    // Read song entries from folder
    RioSongList songs = readSongEntries(folders[folder], folder, card);

    // Build a new name bitmap
    if(!buildTextBitmap(newname, songs[song].Bitmap.Image, songs[song].Bitmap.wNumBlocks))
      throw RIO_E_CREATE_BITMAP_FAILED;

    // Convert BSTR'd name to ASCII string
    memset(songs[song].szName1, 0, 362);
    strncpy(songs[song].szName1, newname, 361);

    // Write song entries back to Rio
    writeSongEntries(folder, songs, card);

    // Wait a bit
    NOP();
    NOP();

    // Get location of song entries    
    DWORD   dwSongBlockOfs = sendCommand(REQ_QUERY_OFFSET_LAST_WRITE, 0x0000, 0x0000);   // no lFlash necessary ?

    // Write to folder
    folders[folder].wOffset = dwSongBlockOfs;
    //Folders[lFolderNum].wFirstFreeEntryOfs += sizeof(SRioFolderEntry);

    // Write folders back to Rio        
    writeFolderEntries(folders, card);

    // Wait a bit..
    NOP();
    NOP();

    // Get location of folder entries..
    DWORD   dwFolderBlockLoc = sendCommand(REQ_QUERY_OFFSET_LAST_WRITE, 0x0000, 0x0000);  // no lFlash necessary ?

    // Tell Rio where the folders are
    sendFolderLocation(dwFolderBlockLoc, folder, card);

    // End folder transfer
    sendCommand(REQ_END_FOLDER_TRANSFERS, 0x0000, card);
  }
  catch (HRESULT e)
  {
    lastError = e;
    return false;
  }
  return true;
}

// OK
bool Rio500::deleteSong(int folder, int song, int card)
{
  lastError = S_OK;
  try
  {
      // Make sure the driver is open
	  if (!IsDriverOpened())
      throw RIO_E_DRIVER_NOT_OPEN;

    // Read folder entries
    RioFolderList folders = readFolderEntries(card);

    // Make sure we have a valid folder
    if (folder >= folders.size())
        throw RIO_E_INVALID_FOLDER;

    // Read song entries
    RioSongList songs = readSongEntries(folders[folder], folder, card);

    // Make sure we have a valid song number
    if (song >= songs.size())
        throw RIO_E_INVALID_SONG;

    // Remove song from the list
    songs.erase(songs.begin() + song);

    // Remove song from Rio.  This works (for some reason)
    sendCommand(REQ_SET_WRITE_ADDRESS, ((folder << 8) | song), card);

    // Write song entries back to Rio
    writeSongEntries(folder, songs, card);

    // Wait a bit..
    NOP();
    NOP();

    // Get location of song entries    
    DWORD   dwSongBlockOfs = sendCommand(REQ_QUERY_OFFSET_LAST_WRITE, 0x0000, 0x0000);  // no lFlash necessary ?

    // Write to folder..
    folders[folder].wOffset = dwSongBlockOfs;
    folders[folder].wFirstFreeEntryOfs -= 0x800;

    // Write folders back to Rio        
    writeFolderEntries(folders,card);

    // Wait a bit..
    NOP();
    NOP();

    // Get location of folders
    DWORD   dwFolderBlockLoc = sendCommand(REQ_QUERY_OFFSET_LAST_WRITE, 0x0000, 0x0000);  // no lFlash necessary ?

    // Tell Rio where the folders are
    sendFolderLocation(dwFolderBlockLoc, folder,card);

    // End folder transfer
    sendCommand(REQ_END_FOLDER_TRANSFERS, 0x0000, card);
  }
  catch (HRESULT e)
  {
    lastError = e;
    return false;
  }

  return true;

}

// OK
RioSongList Rio500::getSongs(int folder, int card)
{
  //---Log << "Getting Songs in folder " << folder << " on card " << card << "\n";
  RioFolderList folders;
  RioSongList list;
  try
  {
    if(!IsDriverOpened()) throw RIO_E_DRIVER_NOT_OPEN;

    folders = readFolderEntries(card);
    if(folder > folders.size())
      throw RIO_E_INVALID_FOLDER;

    list    = readSongEntries(folders[folder], folder, card);
  }
  catch(HRESULT e)
  {
    //---Log << "  Error: " << e << "\n";
    lastError = e;
    return list;
  }
  //---Log << "  Success!\n";
  return list;
}

bool Rio500::putSongs(int folder, RioSongList list, int card)
{
  try
  {
      // Make sure the driver is opened..
	  if (!IsDriverOpened())
      throw RIO_E_DRIVER_NOT_OPEN;

    // Read the folder entries list
    RioFolderList folders = readFolderEntries(card);

    // Make sure we have a valid folder number
    if (folder >= folders.size())
        throw RIO_E_INVALID_FOLDER;

    // Write song entries back to Rio
    writeSongEntries(folder, list, card);

    // Wait a bit
    NOP();
    NOP();

    // Get location of song entries    
    DWORD   dwSongBlockOfs = sendCommand(REQ_QUERY_OFFSET_LAST_WRITE, 0x0000, 0x0000);   // no lFlash necessary ?

    // Write to folder
    folders[folder].wOffset = dwSongBlockOfs;

    // Write folders back to Rio        
    writeFolderEntries(folders, card);

    // Wait a bit..
    NOP();
    NOP();

    // Get location of folder entries..
    DWORD   dwFolderBlockLoc = sendCommand(REQ_QUERY_OFFSET_LAST_WRITE, 0x0000, 0x0000);  // no lFlash necessary ?

    // Tell Rio where the folders are
    sendFolderLocation(dwFolderBlockLoc, folder, card);

    // End folder transfer
    sendCommand(REQ_END_FOLDER_TRANSFERS, 0x0000, card);
  }
  catch (HRESULT e)
  {
    lastError = e;
    return false;
  }
  return true;
}

// OK
void Rio500::getFontName(char *name)
{
  strcpy(name, fontName);
}

// OK
void Rio500::setFontName(char *name)
{
  strcpy(fontName, name);
}

// OK
int  Rio500::getFontSize()
{
  return fontSize;
}

// OK
void Rio500::setFontSize(int size)
{
  fontSize = size;
}

// OK
bool Rio500::getSong(int folder, int song, char *filename, int card, ProgressCallback callback, void *parm)
{
	HANDLE hFile = INVALID_HANDLE_VALUE;
	DWORD myBytesWritten,dwFolderBlockOffset;
	BYTE* pBlock = NULL;
	WORD wOffsetBuffer;
	
	try
  {
        // a memory block for our data		
		pBlock = new BYTE[RIO_HUGEBLOCKSIZE];
		
		// Make sure the driver is open
	  if (!IsDriverOpened())
      throw RIO_E_DRIVER_NOT_OPEN;

    // Read folder entries
    RioFolderList folders = readFolderEntries(card);

    // Make sure we have a valid folder
    if (folder >= folders.size())
        throw RIO_E_INVALID_FOLDER;

    // Read song entries
    RioSongList   songs = readSongEntries(folders[folder], folder, card);

    // Make sure we have a valid song number
    if (song >= songs.size())
        throw RIO_E_INVALID_SONG;

		// Try to open the file
    hFile = CreateFile(filename, GENERIC_WRITE, FILE_SHARE_WRITE, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

    // Make sure it opened successfully
    if (hFile == INVALID_HANDLE_VALUE)
      throw RIO_E_FILE_NOT_FOUND;

		// now we do the trick .. change folder offset with the song offset
		// then write folders to the rio

		// save folder 0's original offset
		wOffsetBuffer = folders[0].wOffset;
		folders[0].wOffset = songs[song].wOffset;

		// write folders
		writeFolderEntries(folders, card);

		// some black magic
		NOP();
		NOP();
		dwFolderBlockOffset = sendCommand(REQ_QUERY_OFFSET_LAST_WRITE, 0, 0);

		// Tell Rio where the root folder block is.
		sendFolderLocation(dwFolderBlockOffset, folders.size(), card);
	
		// black magic again
		sendCommand(REQ_END_FOLDER_TRANSFERS, 0x0, card);

		// setup variables and communication
		DWORD dwReadSizeReq = 0;
		DWORD dwReadSize = 0;
		DWORD dwCount = 0;
		WORD wBlocksToRead = songs[song].dwLength / RIO_HUGEBLOCKSIZE;
		WORD wBlocksAdd = (songs[song].dwLength % RIO_HUGEBLOCKSIZE) / RIO_BLOCKSIZE;
		WORD wRemainder  = (songs[song].dwLength % RIO_HUGEBLOCKSIZE) % RIO_BLOCKSIZE;
		
		sendCommand(REQ_QUERY_FOLDER_BLOCK, 0xff, card);

		// read the file
		for (WORD b = 0; b <= wBlocksToRead; b++) 
		{	
			dwReadSizeReq = (b < wBlocksToRead ? RIO_HUGEBLOCKSIZE : wBlocksAdd*RIO_BLOCKSIZE);
			if (dwReadSizeReq)
			{
				dwReadSize = sendCommand(REQ_READ_FROM_USB, dwReadSizeReq >> 16, dwReadSizeReq & 0xFFFF);

		    if (dwReadSize != dwReadSizeReq)
			    throw RIO_E_COULDNT_ACCEPT_DATA;

				dwCount = 0;
				for (int i=0; i < (b < wBlocksToRead ? RIO_HUGEBLOCKSIZE/RIO_BLOCKSIZE : wBlocksAdd); i++) 
					dwCount += bulkRead(pBlock+i*RIO_BLOCKSIZE,RIO_BLOCKSIZE);

				if (dwCount != dwReadSize)
					throw RIO_E_SHORT_READ;

				WriteFile(hFile,pBlock,dwReadSize,&myBytesWritten,NULL);

        if (wBlocksToRead) 
        {
          if (!callback((b * 100) / wBlocksToRead, parm))
          {
		        // Close the song file
		        if (hFile != INVALID_HANDLE_VALUE)
			        CloseHandle(hFile);

		        // Free song block data
		        if (pBlock)
			        delete pBlock;

		        // restore correct folder offset
		        folders[0].wOffset = wOffsetBuffer;

		        // write folders
		        writeFolderEntries(folders, card);

		        // some black magic
		        sendCommand(REQ_NOP, 0, 0);
		        sendCommand(REQ_NOP, 0, 0);
		        dwFolderBlockOffset = sendCommand(REQ_QUERY_OFFSET_LAST_WRITE, 0, 0);

		        // Tell Rio where the root folder block is.
		        sendFolderLocation(dwFolderBlockOffset, folders.size(), card);

		        // black magic again
		        sendCommand(REQ_END_FOLDER_TRANSFERS, 0x0, card);

            throw RIO_E_USER_ABORTED;
          }
        }

				NOP();
				NOP();
			}
		}

		// read remainder
		if (wRemainder)
		{
			sendCommand(REQ_READ_FROM_USB, 0x0, wRemainder);
			dwCount = bulkRead(pBlock, wRemainder);
			if (dwCount != wRemainder)
				throw RIO_E_SHORT_READ;
			WriteFile(hFile,pBlock,dwCount,&myBytesWritten,NULL);
			NOP();
			NOP();
		}
    callback(100, parm);

		// Close the song file
		if (hFile != INVALID_HANDLE_VALUE)
			CloseHandle(hFile);

		// Free song block data
		if (pBlock)
			delete pBlock;

		// restore correct folder offset
		folders[0].wOffset = wOffsetBuffer;

		// write folders
		writeFolderEntries(folders, card);

		// some black magic
		sendCommand(REQ_NOP, 0, 0);
		sendCommand(REQ_NOP, 0, 0);
		dwFolderBlockOffset = sendCommand(REQ_QUERY_OFFSET_LAST_WRITE, 0, 0);

		// Tell Rio where the root folder block is.
		sendFolderLocation(dwFolderBlockOffset, folders.size(), card);

		// black magic again
		sendCommand(REQ_END_FOLDER_TRANSFERS, 0x0, card);

	}
	catch (HRESULT e)
  {
    lastError = e;
    return false;
  }
  return true;
}


// OK
HRESULT Rio500::getLastError()
{
  return lastError;
}

//******** Protected Member Functions *****************************************

// OK
int Rio500::openDriver()
{
  /*{
    HDEVINFO                hHardwareDevInfo = INVALID_HANDLE_VALUE;
    char                    szRIO500Name[1024];
    DWORD                   i = 0;
    DWORD                   dwNumDevices = 4000;
    bool                    bDone = false;
    bool                    bFoundDevice = false;
    int                     numFound = 0;
    int                     oErrorCount = 0;

    // Loop indefinitely

    DWORD timeStart = timeGetTime();

    DWORD timeNow = timeStart;
    while(true)
    {
      // Get the USB list info stuff
      hHardwareDevInfo = SetupDiGetClassDevs(&GUID_CLASS_RIO500, NULL, NULL, (DIGCF_PRESENT | DIGCF_INTERFACEDEVICE));
      if(hHardwareDevInfo == INVALID_HANDLE_VALUE) goto OpenDriverTryTwo;

      SP_INTERFACE_DEVICE_DATA   DeviceInfoData;
      ZeroMemory(&DeviceInfoData, sizeof(SP_INTERFACE_DEVICE_DATA));
      DeviceInfoData.cbSize = sizeof(SP_INTERFACE_DEVICE_DATA);

      // Set the "found device" flag to false
      bFoundDevice = false;

      // Now, loop starting from the previously found entry
      for(i = numFound; true; i++)
      {
        // If we can get the USB info
        if (SetupDiEnumDeviceInterfaces(hHardwareDevInfo, 0, &GUID_CLASS_RIO500, i, &DeviceInfoData)) 
        {
  		    PSP_INTERFACE_DEVICE_DETAIL_DATA pFunctionClassDeviceData = NULL;
          DWORD                            dwPredictedLength = 0;
          DWORD                            dwRequiredLength = 0;
          // AND if we haven't already found a device
        
          if(!bFoundDevice)
          {
            // Get the USB device details
            SetupDiGetInterfaceDeviceDetail(hHardwareDevInfo, &DeviceInfoData, NULL, 0, &dwRequiredLength, NULL);

            dwPredictedLength = dwRequiredLength;
            pFunctionClassDeviceData = (PSP_INTERFACE_DEVICE_DETAIL_DATA)malloc(dwPredictedLength);
            pFunctionClassDeviceData->cbSize = sizeof(SP_INTERFACE_DEVICE_DETAIL_DATA);

            szRIO500Name[0] = 0;
            if (SetupDiGetInterfaceDeviceDetail(hHardwareDevInfo, &DeviceInfoData, pFunctionClassDeviceData, dwPredictedLength, &dwRequiredLength, NULL))
            {
              strcpy(szRIO500Name, pFunctionClassDeviceData->DevicePath);
            }

            // Set the "we've found one" flags
            bFoundDevice = true;
            numFound = i;
          }
        }
        // Otherwise, if there are no more entries
        else if(GetLastError() == ERROR_NO_MORE_ITEMS)
        {
          // If we haven't found a device, return an error
          if(!bFoundDevice) 
          {
            goto OpenDriverTryTwo;
          }
          // Break the loop
          break;
        }
        else
        {
          oErrorCount++; // Allow 5 other errors
          if(oErrorCount >= 5) goto OpenDriverTryTwo;
        }
      }

      // Destroy our info list
      SetupDiDestroyDeviceInfoList(hHardwareDevInfo);

      // Attempt to create the interface
      rio500 = CreateFile(
        szRIO500Name,                         // Name of device driver's file object
        GENERIC_READ | GENERIC_WRITE,         // Read/write access
        FILE_SHARE_READ | FILE_SHARE_WRITE,   // Share read/write
        NULL,                                 // No SECURITY_ATTRIBUTES structure
        OPEN_EXISTING,                        // No special create flags
        0,                                    // No special attributes                          
        NULL);                                // No template file

      // If that was successful..
      if (rio500 != INVALID_HANDLE_VALUE) 
      {
        rio500Read = rio500Write = rio500;

        Sleep(50);
        // Try to start Rio Comm...if successful...

        //IORequest(DIR_STANDARD, REQ_GET_DESCRIPTOR, DESC_DEVICE, 0, sizeof(UsbDevice), &deviceInfo);
        //char stringTemp[2000];
        //wsprintf(stringTemp, "%04x - %04x", deviceInfo.idVendor, deviceInfo.idProduct);
        //MessageBox(NULL, stringTemp, "Test", MB_OK);
        //if(deviceInfo.idVendor == 
        if(startComm())
        {
          return 0; // Return success
        }
        CloseHandle(rio500);
        // Not successful, so clear our handles.
        rio500Read = rio500Write = rio500 = INVALID_HANDLE_VALUE;
      }
      // Keep looping.
      numFound++;
    }
  }

OpenDriverTryTwo:;

  // First , we need to initialize all of the loop variables
  {
    HDEVINFO                 devInfo;
    SP_INTERFACE_DEVICE_DATA devData;
    PUSB_DEVICE_DESCRIPTOR   usbDeviceInst;
    int                      i              = 0;
    PUSB_DEVICE_DESCRIPTOR  *usbDevices     = &usbDeviceInst;
    bool                     done           = false;
    ULONG                    numDevices     = 4;

    *usbDevices = NULL;

    devInfo = SetupDiGetClassDevs(&GUID_CLASS_RIO500, NULL, NULL, (DIGCF_PRESENT | DIGCF_INTERFACEDEVICE));
    if(devInfo == INVALID_HANDLE_VALUE) goto OpenDriverTryTwo;

    rio500 = INVALID_HANDLE_VALUE;

    memset(&devData, 0, sizeof(SP_INTERFACE_DEVICE_DATA));
    devData.cbSize = sizeof(SP_INTERFACE_DEVICE_DATA);


    // Get them ... ALL of them.
    while(!done)
    {
      // Need more devices, obviously.
      numDevices *= 2;

      // (Re)allocate as necessary
      if(*usbDevices)
        *usbDevices = (PUSB_DEVICE_DESCRIPTOR)realloc(*usbDevices, (numDevices * sizeof(USB_DEVICE_DESCRIPTOR)));
      else
        *usbDevices = (PUSB_DEVICE_DESCRIPTOR)calloc(numDevices, sizeof(USB_DEVICE_DESCRIPTOR));

      // Did it go alright?
      if(*usbDevices == NULL)
      {
        SetupDiDestroyDeviceInfoList(devInfo);
        goto OpenDriverTryThree;
      }

      // Set the USB Device pointer to the right place.
      usbDeviceInst = *usbDevices + i;

      for(;i < numDevices; i++)
      {
        //Enumerate the devices
        if(SetupDiEnumDeviceInterfaces(devInfo, 0, &GUID_CLASS_RIO500, i, &devData))
        {
          // Open the device
          //************************************************************************************

          PSP_INTERFACE_DEVICE_DETAIL_DATA  functionClassDeviceData = NULL;
          ULONG                             predictedLength         = 0;
          ULONG                             requiredLength          = 0;

          SetupDiGetInterfaceDeviceDetail(devInfo, &devData, 0, 0, &requiredLength, 0);

          predictedLength = requiredLength;

          functionClassDeviceData = (PSP_INTERFACE_DEVICE_DETAIL_DATA)malloc(predictedLength);

          memset(functionClassDeviceData, 0, predictedLength);
          functionClassDeviceData->cbSize = sizeof(SP_INTERFACE_DEVICE_DETAIL_DATA);

          if(!SetupDiGetInterfaceDeviceDetail(devInfo, &devData, functionClassDeviceData, predictedLength, &requiredLength, 0))
          {
            free(functionClassDeviceData);
            continue; // Go on with the for loop
          }

          rio500 = CreateFile(functionClassDeviceData->DevicePath, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, 0, OPEN_EXISTING, 0, 0);

          free(functionClassDeviceData);

          //************************************************************************************

          // If it all opened alright...
          if(rio500 != INVALID_HANDLE_VALUE)
          {
            if(startComm())
            {
              done = true;
              break;
            }
            // Did not start okay.
            CloseHandle(rio500);
            rio500 = INVALID_HANDLE_VALUE;

          }
        }
        else // Otherwise...
        {
          //if(GetLastError() == ERROR_NO_MORE_ITEMS) // Didn't find anything, it looks like.
          //{
            done = true;
            break;
          //}
        }
      }
    }

    numDevices = i;


    //  we should make sure it opens and return a spec. error  --tk
    if (rio500 == INVALID_HANDLE_VALUE) 
      goto OpenDriverTryThree;

    rio500Read = rio500Write = rio500;

    // all ok !  --tk
    return 0;
  }

OpenDriverTryThree:;*/
  {
    //---Log << "    Opening driver...\n";
    SP_DEVICE_INTERFACE_DATA devInfoData;
    bool lastDevice = false;
    int memberIndex = 0;
    bool myDeviceDetected = false;
    LONG result;
    DWORD length, required;

    HDEVINFO devInfo;
  
    //---Log << "    SetupDiGetClassDevs!\n";
    devInfo = SetupDiGetClassDevs(&GUID_CLASS_RIO500, 0, 0, DIGCF_PRESENT|DIGCF_DEVICEINTERFACE);

    devInfoData.cbSize = sizeof(devInfoData);
  
    //---Log << "    Entering Do Loop!\n";
    do
    {
      myDeviceDetected = false;
    
      //---Log <<"      SetupDiEnumDeviceInterface!\n";
      result = SetupDiEnumDeviceInterfaces(devInfo, 0, &GUID_CLASS_RIO500, memberIndex, &devInfoData);
    
      if(result != 0)
      {
        //---Log << "        Result != 0!\n";
		    PSP_INTERFACE_DEVICE_DETAIL_DATA detailData = NULL;

        //---Log << "        SetupDiGetDeviceInterfaceDetail!\n";
        result = SetupDiGetDeviceInterfaceDetail(devInfo, &devInfoData, 0, 0, &required, 0);
        length = required;

        //---Log << "        Allocating Detail Data\n";
        detailData = (PSP_DEVICE_INTERFACE_DETAIL_DATA)malloc(length);
        detailData->cbSize = sizeof(SP_INTERFACE_DEVICE_DETAIL_DATA);

        //---Log << "        SetupDiGetDeviceInterfaceDetail (take 2)\n";
        result = SetupDiGetDeviceInterfaceDetail(devInfo, &devInfoData, detailData, length, &required, 0);
        if(result != 0)
        {
          //---Log << "          Creating file!\n";
          rio500 = CreateFile(detailData->DevicePath, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE, 0, OPEN_EXISTING, 0, 0);

          if(rio500 != INVALID_HANDLE_VALUE)
          {
            //---Log << "            Created!\n";
            rio500Read = rio500Write = rio500;

            if(startComm())
            {
              //---Log << "              Start Comm Success!\n";
              myDeviceDetected = true;
            }
            else
            {
              //---Log << "              Failed.  Not a Rio!\n";
              CloseHandle(rio500);
              rio500Read = rio500Write = rio500 = INVALID_HANDLE_VALUE;
            }
          }
          //---Log << "        Freeing detailData!\n";
          free(detailData);
        }
      }
      else
      {
        //---Log << "        Last device!\n";
        char string[2000];
        FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, 0, GetLastError(), 0, string, 2000, 0);
        //---Log << "        Message: " << string << "\n";
        lastDevice = true;
      }
      memberIndex++;
    }
    while(!lastDevice && !myDeviceDetected);

    //---Log << "    Loop Ended!\n";

    //---Log << "    Destroying device list!\n";
    SetupDiDestroyDeviceInfoList(devInfo);

    //---Log << "    Returning!\n";
    if(myDeviceDetected)
      return 0;
  }
  return 1;
}

int Rio500::openDriverRaw()
{
	HKEY myRegKey;
	DWORD dwIndex = 0;
	TCHAR    szBuffer[1024];
    BOOL bDone = false;

	ZeroMemory(szBuffer, 1024);

	RegOpenKey(HKEY_LOCAL_MACHINE,"System",&myRegKey);
	RegOpenKey(myRegKey,"CurrentControlSet",&myRegKey);
	RegOpenKey(myRegKey,"Control",&myRegKey);
	RegOpenKey(myRegKey,"DeviceClasses",&myRegKey);
	if(RegOpenKey(myRegKey,"{a5dcbf10-6530-11d2-901f-00c04fb951ed}",&myRegKey) == ERROR_SUCCESS)
  {
	  while(!(RegEnumKey(myRegKey,dwIndex,szBuffer,sizeof(szBuffer)) == ERROR_NO_MORE_ITEMS))
	  {
		  
		  int i = 0;
		  for (i = 0;i < 1024; i++) {
			  if (szBuffer[i] == 0x23) szBuffer[i] = 0x5C;

		  };
		  szBuffer[20]=0x23;
		  
		  // Open the device driver
		  rio500 = CreateFile(
		  szBuffer,							  // Name of device driver's file object
		  GENERIC_READ | GENERIC_WRITE,         // Read/write access
		  FILE_SHARE_READ | FILE_SHARE_WRITE,   // Share read/write
		  NULL,                                 // No SECURITY_ATTRIBUTES structure
		  OPEN_EXISTING,                        // No special create flags
		  0,                                    // No special attributes                          
		  NULL);                                // No template file
	  
		  if (rio500 != INVALID_HANDLE_VALUE) 
		  {
			  
			  rio500Read = rio500Write = rio500;
			  
			  // make sure it is the rio .. :)
        Sleep(50);
			  if (startComm())
			  {
				  // all ok !
				  return 0;
			  }
			  else
			  {
				  CloseHandle(rio500);
			  };
		  };

		  dwIndex ++;
	  };
  }

	// ok now we try the older CLSID !
	
	ZeroMemory(szBuffer, 1024);
	dwIndex = 0;
	bDone = false;

	RegOpenKey(HKEY_LOCAL_MACHINE,"System",&myRegKey);
	RegOpenKey(myRegKey,"CurrentControlSet",&myRegKey);
	RegOpenKey(myRegKey,"Control",&myRegKey);
	RegOpenKey(myRegKey,"DeviceClasses",&myRegKey);
	if(RegOpenKey(myRegKey,"{4e183ba0-e87e-11d2-bde0-00902743f01d}",&myRegKey) == ERROR_SUCCESS)
  {
	  while(!(RegEnumKey(myRegKey,dwIndex,szBuffer,sizeof(szBuffer)) == ERROR_NO_MORE_ITEMS))
	  {
		  
		  int i = 0;
		  for (i = 0;i < 1024; i++) {
			  if (szBuffer[i] == 0x23) szBuffer[i] = 0x5C;

		  };
		  szBuffer[20]=0x23;
		  
		  // Open the device driver
		  rio500 = CreateFile(
		  szBuffer,							  // Name of device driver's file object
		  GENERIC_READ | GENERIC_WRITE,         // Read/write access
		  FILE_SHARE_READ | FILE_SHARE_WRITE,   // Share read/write
		  NULL,                                 // No SECURITY_ATTRIBUTES structure
		  OPEN_EXISTING,                        // No special create flags
		  0,                                    // No special attributes                          
		  NULL);                                // No template file
	  
		  if (rio500 != INVALID_HANDLE_VALUE) 
		  {
			  
			  rio500Read = rio500Write = rio500;
			  
			  // make sure it is the rio .. :)
        Sleep(50);
			  if (startComm())
			  {
				  // all ok !
				  return 0;
			  }
			  else
			  {
				  CloseHandle(rio500);
			  };
		  };

		  dwIndex ++;
	  }
  }
	// nothing worked ...
	return 1;
};


// OK
void Rio500::closeDriver()
{
  if(rio500 != INVALID_HANDLE_VALUE)
  {
    CloseHandle(rio500);
    rio500 = INVALID_HANDLE_VALUE;
  }
}


// OK
bool Rio500::startComm()
{
  //---Log << "Starting Comm!\n";
  if(sendCommand(REQ_START_USB_COMM, 0x0000, 0x0000) == RSTATUS_ERROR) return false;
  NOP();
  NOP();

  return true;
}

// OK
bool Rio500::endComm()
{
  //---Log << "Ending Comm!\n";
  if(sendCommand(REQ_END_USB_COMM, 0x0000, 0x0000) == RSTATUS_ERROR) return false;
  NOP();
  NOP();

  return true;
}

// OK
DWORD Rio500::sendCommand(RIO_REQUEST Req, WORD wValue, WORD wIndex)
{
  //---Log << "  Send Command!\n";
  DWORD dwStatus;

  // Send the request..
  if (!IORequest(DIR_IN, Req, wValue, wIndex, sizeof(DWORD), &dwStatus))
      return RSTATUS_ERROR;

  // Return the result..
  //---Log << "    Returning!\n";
  return dwStatus;
}

// OK
bool Rio500::IORequest(RIO_DIRECTION Dir, RIO_REQUEST Req, WORD wValue, WORD wIndex, DWORD dwLen, void* pData)
{
  DWORD       dwBytesReturned = 0;

  // Diamond's custom structure for a USB request..
  struct
  {
      BYTE    byRequestType;
      BYTE    byRequest;
      WORD    wValue;
      WORD    wIndex;
      WORD    wLen;
      void*   pData;

  } IOBlock;
  
  

  // Fill out the IO block..
  switch(Dir)
  {
    case DIR_IN:
      IOBlock.byRequestType = 0xC0;
      break;
    case DIR_STANDARD:
      IOBlock.byRequestType = 0x80;
      break;
    default:
      IOBlock.byRequestType = 0x40;
      break;
  }
  //IOBlock.byRequestType = (Dir == DIR_IN) ? 0xC0 : 0x40;
  //if(Dir == DIR_STANDARD) IOBlock.byRequestType = 0x80;
  IOBlock.byRequest = Req;
  IOBlock.wValue = wValue;
  IOBlock.wIndex = wIndex;
  IOBlock.wLen = (WORD)dwLen;
  IOBlock.pData = pData;

  // Have a chat with the device driver..
  if (!DeviceIoControl(rio500, CTL_CODE(FILE_DEVICE_UNKNOWN, 3, METHOD_BUFFERED, FILE_ANY_ACCESS), &IOBlock, sizeof(IOBlock), pData, dwLen, &dwBytesReturned, NULL))
	{
		// got an error .. we wait 50 msecs and retry
	
		Sleep(50);
		
		if (!DeviceIoControl(rio500, CTL_CODE(FILE_DEVICE_UNKNOWN, 3, METHOD_BUFFERED, FILE_ANY_ACCESS), &IOBlock, sizeof(IOBlock), pData, dwLen, &dwBytesReturned, NULL)) 
    {
			return false;
    }
	}
    
	// We're happy...
    return true;
}

// OK
bool Rio500::buildTextBitmap(char *bsText, BYTE* pBitmapData, WORD& wBlocks)
{
  bool                bReturn  = true;
  HBITMAP             hBMP     = NULL;
  HDC                 hDC      = NULL;
  HBITMAP             hOldBMP  = NULL;
  int                 nTextLen = strlen(bsText);
  HFONT               hFont    = NULL;
  HFONT               hOldFont = NULL;
  BITMAPINFO*         pBMI     = NULL;
  RECT r;
  SIZE TextExtent;

  try
  {

    // Make a bitmap..
    if ((hBMP = CreateBitmap(768, 16, 1, 1, NULL)) == NULL)
        throw false;

    // Create a DC to draw in the bitmap..
    if ((hDC = CreateCompatibleDC(NULL)) == NULL)
        throw false;

    // Put the bitmap in the DC..
    if ((hOldBMP = (HBITMAP)SelectObject(hDC, hBMP)) == NULL)
        throw false;

    // The size of the bitmap..
    SetRect(&r, 0, 0, 768, 16);

    // Make it all white..
    if (!FillRect(hDC, &r, (HBRUSH)GetStockObject(WHITE_BRUSH)))
        throw false;

    // Calculate the font height..        
    int nHeight = -MulDiv(fontSize, GetDeviceCaps(hDC, LOGPIXELSY), 72);

    // Make the font..
    hFont = CreateFont(nHeight, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, fontName);

    // Did it work?
    if (hFont == NULL)
      throw false;

    // Put the font in the DC
    if ((hOldFont = (HFONT)SelectObject(hDC, hFont)) == NULL)
      throw false;

    // Get the extent of the text
    if (!GetTextExtentPoint32(hDC, bsText, nTextLen, &TextExtent))
      throw false;

    // divide by 8 to get number of blocks..
    wBlocks = TextExtent.cx / 8;

    // Add one if we have any extra bits..
    if (TextExtent.cx % 8)
      wBlocks++;

    // Black text please..        
    SetTextColor(hDC, RGB(0x00, 0x00, 0x00));

    // Draw the text
    if (!TextOut(hDC, 0, 0, bsText, nTextLen))
      throw false;

    // Setup a temporary bitmap to get the DIB bits too..
    BYTE    TempBitmap[RIO_BITMAPSIZE];
    DWORD   dwBMISize = sizeof(BITMAPINFO) + sizeof(RGBQUAD) * 1;

    // setup a nice new bitmap info struct..
    pBMI = (BITMAPINFO*)malloc(dwBMISize);

    pBMI->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    pBMI->bmiHeader.biHeight = -16;
    pBMI->bmiHeader.biWidth = 768;
    pBMI->bmiHeader.biPlanes = 1;
    pBMI->bmiHeader.biBitCount = 1;
    pBMI->bmiHeader.biCompression = BI_RGB;

    // Get DIB bits..
    BOOL bOk = GetDIBits(hDC, hBMP, 0, 16, TempBitmap, pBMI, DIB_RGB_COLORS);

    // This gets a little weird...  we have to transmorgify the bitmap into 8-pixel wide
    // blocks, instead of just giving it to the Rio straight..  While we're at it, we 
    // invert the bits so it shows up properly on the Rio's display.  For those of you 
    // playing the home game, you're asking yourselves "Why didn't he just render white
    // text on a black background?"  Because it didn't work..  for some reason.  I always
    // just got a big black bitmap - which was pretty boring.  I'm not sure what the deal
    // is, but because I A) have to transmorgify it anyway, it doesn't matter and B) didn't
    // want to screw with it a lot.  2 bonus points if you now the origin of "transmogrify"
    BYTE* p = pBitmapData;
    ZeroMemory(pBitmapData, RIO_BITMAPSIZE);

    // Transmorgify
    for (WORD b = 0; b < wBlocks; b++)
    {
        for (WORD y = 0; y < 16; y++, p++)
            *p = ~TempBitmap[(y*96)+b];
    }
  }
  catch (bool e)
  {
    bReturn = e;
  }

  // Free bitmap info struct
  if (pBMI)
    free(pBMI);

  // Free font
  if (hOldFont && hDC && hFont)
  {
    SelectObject(hDC, hOldFont);
    DeleteObject(hFont);
  }

  // Free bitmap
  if (hOldBMP && hDC && hBMP)
  {
    SelectObject(hDC, hOldBMP);
    DeleteObject(hBMP);
  }

  // Free DC
  if (hDC)
    DeleteDC(hDC);

  return bReturn;
}

// OK
bool Rio500::isFirstFolder(long lFlash)
{
  DWORD dwResult;

  // Get the number of folders..
  if ((dwResult = sendCommand(REQ_QUERY_NUM_FOLDER_BLOCKS, 0xFF00, lFlash)) > 0)
    return false;

  // Wait a bit.
  NOP();
  NOP();

  // Try again, just to make sure.. (for some reason)
  if ((dwResult = sendCommand(REQ_QUERY_NUM_FOLDER_BLOCKS, 0xFF00, lFlash)) > 0)
    return false;

  // Wait a bit..
  NOP();
  NOP();

  // Okay, we're done.
  sendCommand(REQ_END_FOLDER_TRANSFERS, 0xFF00, lFlash);

  return true;
}

// OK
RioFolderList Rio500::readFolderEntries(long lFlash)
{
  RioFolderList       Folders;
  BYTE*               pBlock = NULL;
  DWORD               dwTotalRead = 0;
  DWORD               dwFolderCount = 0;
  DWORD               dwComStatus = 0;
  RioFolderEntry*     pFE;

  // The mysterious read command...
  if (!sendReadCommand(0xFF00, 1, lFlash))
      return Folders;

  // Allocate memory for a block of data..
  BYTE*   pFolderBlock = new BYTE[RIO_BLOCKSIZE];

  // Read some data.
  dwTotalRead = bulkRead(pFolderBlock, RIO_BLOCKSIZE);

  // Make sure we got it okay.
  if (dwTotalRead != RIO_BLOCKSIZE)
  {
    delete pFolderBlock;
    //throw RIO_E_COULDNT_ACCEPT_DATA;
    return Folders;
  }

  // cast the pointer to somethign we can deal with..
  pFE = (RioFolderEntry*)pFolderBlock;

  // Loop through folders, adding them to the collection
  while (pFE->wOffset != 0xFFFF && dwFolderCount < 8)
  {
    Folders.push_back(*pFE);
    dwFolderCount++;
    pFE++;
  }

  // Free memory for folder block
  delete pFolderBlock;

  return Folders;
}

// OK
bool Rio500::allocateEmptyFolderBlock(DWORD dwNumFolders, long lFlash)
{
  DWORD   dwAddress = dwNumFolders;
  BYTE*   pEmptyBlock = new BYTE[RIO_BLOCKSIZE];

  // Munge address.  for some reason..
  dwAddress <<= 8;
  dwAddress |= 0x000000FF;
  dwAddress &= 0x00000FFF;

  // "format" the empty block
  clearBlock(pEmptyBlock);

  // Tell the Rio it's about to get something..
  bool bRetVal = sendWriteCommand(dwAddress, 1, lFlash);
  
  if (bRetVal)
  {
    // Wait a bit..    
    NOP();
    NOP();

    // Blast out the empty folder block
    bRetVal = bulkWrite(pEmptyBlock, RIO_BLOCKSIZE) == RIO_BLOCKSIZE;
  }
  // Clean up after ourselves..
  delete pEmptyBlock;

  return bRetVal;
}

// OK
void Rio500::clearBlock(BYTE *pBlock)
{
  BYTE* p;

  // Bzzzt..
  ZeroMemory(pBlock, RIO_BLOCKSIZE);

  // Initialize folder entries to signal "no folder here"..
  p = pBlock;
  if (pBlock)
  {
    while ((p - pBlock) < RIO_BLOCKSIZE)
    {
      *p     = 0xFF;
      *(p+1) = 0xFF;
      p += sizeof(RioFolderEntry);
    }
  }
}

// OK
bool Rio500::sendWriteCommand(DWORD dwAddress, DWORD dwNumBlocks, long lFlash)
{
  // Calculate length we'll be sending..
  DWORD   dwLength = dwNumBlocks * RIO_BLOCKSIZE;
  WORD    wWriteBlocks = WORD(dwLength / RIO_BIGBLOCKSIZE);
  WORD    wWriteRemainder = WORD(dwLength % RIO_BIGBLOCKSIZE);
  DWORD   dwWriteStatus = 0;

  // Give Rio the address we'd like to write to..
  dwWriteStatus = sendCommand(REQ_SET_WRITE_ADDRESS, dwAddress, lFlash);

  // Who knows...
  if (!(dwWriteStatus = sendCommand(REQ_MYSTERY, 0xFFFF, lFlash)))
    return false;

  // Tell Rio to open wide...  
  if (!(dwWriteStatus = sendCommand(REQ_WRITE_TO_USB, wWriteBlocks, wWriteRemainder)))
    return false;

  return true;
}

// OK
DWORD Rio500::bulkWrite(void* pBlock, DWORD dwNumBytes)
{
  BYTE*   p = (BYTE*)pBlock;
  DWORD   dwCount = 0;
  DWORD   dwBytesLeft = dwNumBytes;
  DWORD   dwBytesWritten = 0;

  do
  {
    BOOL bRet;

    // Blast away...
    bRet = WriteFile(rio500Write, p, dwBytesLeft, &dwCount, NULL);
    // Make it worked, and calculate how much more to write..
    if (dwCount > 0)
    {
      p += dwCount;
      dwBytesWritten += dwCount;
      dwBytesLeft -= dwCount;
    }
  }
  while (dwBytesLeft > 0 && dwCount  > 0);

  // Return number of bytes written
  return dwBytesWritten;
}

// OK
bool Rio500::writeFolderEntries(RioFolderList& Folders, long lFlash)
{
  // Do some calculations up front to figure the number of blocks we'll need..
  DWORD   dwListLen = Folders.size();
  DWORD   dwNumBlocks = dwListLen >> 3;
  DWORD   dwNumFolders = 0;

  // Check for remainder..    
  if (dwListLen & 0x00000007)
    dwNumBlocks++;

  // UPDATE BY JJ:-------------------------------------//
  // MAJOR fix...If there are NO folders, then I special
  // cased the scenario as needing one block...otherwise,
  // crazy stuff happened. (Like, it hanging)
  if(!dwListLen) 
  {
    dwNumBlocks = 1;
    // Tell Rio we're going to write to it..
    sendWriteCommand(0xFF00, dwNumBlocks, lFlash);
    BYTE* pFolderBlock = new BYTE[RIO_BLOCKSIZE];

    // Clear folder block..
    clearBlock(pFolderBlock);
    
    // Write the block
    bool bRetVal = bulkWrite(pFolderBlock, RIO_BLOCKSIZE) == RIO_BLOCKSIZE;
    delete pFolderBlock;
    return bRetVal;
  }
  // END UPDATE ---------------------------------------//


  // Tell Rio we're going to write to it..
  sendWriteCommand(0xFF00, dwNumBlocks, lFlash);

  // Wait a bit..
  NOP();
  NOP();

  // Allocate temporary buffer
  BYTE*               pFolderBlock = new BYTE[RIO_BLOCKSIZE];
  RioFolderEntry*    pFE = (RioFolderEntry*)pFolderBlock;

  // Clear folder block..
  clearBlock(pFolderBlock);

  // Copy data from collection into block
  for (RioFolderList::iterator it = Folders.begin(); it != Folders.end() && dwNumFolders < 8; it++, dwNumFolders++, pFE++)
    memcpy(pFE, it, sizeof(RioFolderEntry));

  // Write the block
  bool bRetVal = bulkWrite(pFolderBlock, RIO_BLOCKSIZE) == RIO_BLOCKSIZE;

  // Delete temporary data
  delete pFolderBlock;
  return bRetVal;

}

// OK
bool Rio500::sendFolderLocation(DWORD dwOffset, DWORD dwFolderNum, long lFlash)
{
  RioFolderLocation  Location;

  // Bzzt..    
  ZeroMemory(&Location, sizeof(RioFolderLocation));

  // Setup structure..
  Location.wOffset = (WORD)dwOffset;
  Location.wBytes = (WORD)RIO_BLOCKSIZE;
  Location.wFolderNum = (WORD)dwFolderNum; 

  // The only DIR_OUT IO request we need..
  return IORequest(DIR_OUT, REQ_SEND_FOLDER_LOCATION, 0x0000, lFlash, sizeof(RioFolderLocation), &Location);
}

// OK
bool  Rio500::sendReadCommand(DWORD dwAddress, DWORD dwNumBlocks, long lFlash)
{
  //---Log << "  Sending Read Command!\n";
  // Figure how much data we want..
  DWORD   dwLength = dwNumBlocks * RIO_BLOCKSIZE;
  WORD    wReadBlocks = WORD(dwLength / RIO_BIGBLOCKSIZE);
  WORD    wReadRemainder = WORD(dwLength % RIO_BIGBLOCKSIZE);
  DWORD   dwReadStatus = 0;

  // Setup Rio for reading...
  if (!(dwReadStatus = sendCommand(REQ_QUERY_NUM_FOLDER_BLOCKS, (WORD)dwAddress, lFlash)))
    return false;
  
  // Setup Rio for reading...
  if (!(dwReadStatus = sendCommand(REQ_QUERY_FOLDER_BLOCK, (WORD)dwAddress, lFlash)))
    return false;

  // Tell Rio how much data we're going to read in 64k chunks and the remainder..
  if (!(dwReadStatus = sendCommand(REQ_READ_FROM_USB, wReadBlocks, wReadRemainder)))
    return false;

  return true;

}

// OK
DWORD Rio500::bulkRead(void* pBlock, DWORD dwNumBytes)
{
  //---Log << "  Bulk Reading " << dwNumBytes << " bytes\n";
  DWORD   dwTotalRead = 0;
  DWORD   dwCount = 0;
  DWORD   dwBytesLeft = dwNumBytes;
  BYTE*   p = (BYTE*)pBlock;

  do
  {
    // Do the read - make sure it worked..
    if (!(ReadFile(rio500Read, p, dwBytesLeft, &dwCount, NULL)))
    {
      DWORD dwError = GetLastError();
    }

    // Figure out if we need to read more..
    if (dwCount > 0)
    {
      dwTotalRead += dwCount;
      dwBytesLeft -= dwCount;
      p += dwCount;
    }
    // Read more if necessary..
  }
  while (dwTotalRead < dwNumBytes && dwCount > 0);

  // Return total bytes read
  //---Log << "    Returning!\n";
  return dwTotalRead;

}

// OK
RioSongList Rio500::readSongEntries(RioFolderEntry& Folder, DWORD dwFolderNum, long lFlash)
{
  // Calculate number of blocks we'll need..
  RioSongList     Songs;
  DWORD           dwNumBlocks = Folder.wFirstFreeEntryOfs / RIO_BLOCKSIZE;
  DWORD           dwCount = 8 * dwNumBlocks;
  //---Log << "    Reading Song Entries: " << dwNumBlocks << " - " << dwCount << "\n";

  // Check remainder..
  if ((Folder.wFirstFreeEntryOfs % RIO_BLOCKSIZE) > 0)
      dwNumBlocks++;
  //---Log << "    New Num Blocks: " << dwNumBlocks << "\n";

  // Add in remainder..
  dwCount += (Folder.wFirstFreeEntryOfs % RIO_BLOCKSIZE) / sizeof(RioFolderEntry);
  //---Log << "    Count: " << dwCount << "\n";

  // No blocks?  No songs!
  if (((int)dwNumBlocks) <= 0) // UPDATE BY JJ: Switched this to recognize numbers LESS than 0 as bad, too.
      return Songs;

  // Calculate folder address..
  DWORD   dwAddress = dwFolderNum;

  dwAddress <<= 8;
  dwAddress |= 0x000000FF;
  dwAddress &= 0x00000FFF;
  //---Log << "    dwAddress: " << dwAddress << "\n";

  // Setup buffers for reading song entry blocks..
  DWORD   dwSize = dwNumBlocks * RIO_BLOCKSIZE;
  BYTE*   pSongBlock = new BYTE[dwSize];

  // Tell Rio we're going to read..
  sendReadCommand(dwAddress, dwNumBlocks, lFlash);

  // Do the read..
  dwSize = RIO_BLOCKSIZE;
  //---Log << "    Looping through blocks\n";
  for(int looper = 0; looper < dwNumBlocks; looper++)
  {
    //---Log << "      Block "  << looper << "\n";
    DWORD   dwTotalRead = bulkRead(pSongBlock, dwSize);

    //---Log << "      Total Read: " << dwTotalRead << " (" << dwSize << ")\n";

    // If we didn't get what we expect, stop.
    if (dwTotalRead != dwSize)
    {
      delete pSongBlock;
      return Songs;
    }

    // Step through song entries, adding them to the collection..
    //---Log << "      Getting pointer to song block\n";
    RioSongEntry*  pSE = (RioSongEntry*)pSongBlock;

    int count2 = 8;
    //---Log << "      Looping through songs in block\n";
    while (dwCount && pSE->wOffset != 0xFFFF && count2)
    {
      //---Log << "        Pushing onto stack: " << count2 << "\n";
      Songs.push_back(*pSE);

      //---Log << "        Incrementing\n";
      pSE++;
      dwCount--;
      count2--;
    }
    //---Log << "      End while loop!\n";
  }
  // Clean up..
  delete pSongBlock;

  //---Log << "    We're done here, returning\n";
  return Songs;
}

// OK
bool Rio500::writeSongEntries(DWORD dwFolderNum, RioSongList& Songs, long lFlash)
{
  DWORD   dwAddress = dwFolderNum;
  BYTE*   pBlock = NULL;
  bool    bRetVal = true;

  try
  {
    // Munge address..
    dwAddress <<= 8;
    dwAddress |= 0x000000FF;
    dwAddress &= 0x00000FFF;

    // Calculate number of blocks we'll need..
    DWORD   dwNumBlocks = Songs.size() >> 3;

    // Handle remainder..
    if (Songs.size() & 0x7)
        dwNumBlocks++;

    // Tell Rio we need to write to it...
    sendWriteCommand(dwAddress, dwNumBlocks, lFlash);

    // Allocate temporary block and clear it.
    pBlock = new BYTE[RIO_BLOCKSIZE];
    clearBlock(pBlock);

    RioSongEntry*   pSE = (RioSongEntry*)pBlock;
    DWORD           dwCount = 0;

    // Walk STL collection writing song entries..
    for (RioSongList::iterator it = Songs.begin(); it != Songs.end(); it++)
    {
      // Copy song entry to block..
      memcpy(pSE, it, sizeof(RioSongEntry));
      dwCount++;
      pSE++;

      // Only 8 entries per block.. 
      if (dwCount == 8)
      {
        // Write out full song entry block
        if (bulkWrite(pBlock, RIO_BLOCKSIZE) != RIO_BLOCKSIZE)
          throw false;

        // Bzzt..
        dwCount = 0;
        clearBlock(pBlock);
        pSE = (RioSongEntry*)pBlock;
      }
    }

    // Write remaining song entry blocks..
    if (dwCount)
    {
      if (bulkWrite(pBlock, RIO_BLOCKSIZE) != RIO_BLOCKSIZE)
        throw false;
    }
  }
  catch (bool e)
  {
    bRetVal = e;
  }

  // Clean up
  if (pBlock)
    delete pBlock;

  return bRetVal;
}
