const HRESULT RIO_E_COULDNT_OPEN_DEVICE      = -1*MAKE_HRESULT(SEVERITY_SUCCESS,FACILITY_ITF,0x6000);
const HRESULT RIO_E_COMM_START_ERROR         = -1*MAKE_HRESULT(SEVERITY_SUCCESS,FACILITY_ITF,0x6001);
const HRESULT RIO_E_DRIVER_NOT_OPEN          = -1*MAKE_HRESULT(SEVERITY_SUCCESS,FACILITY_ITF,0x6002);
const HRESULT RIO_E_FORMAT_FAILED            = -1*MAKE_HRESULT(SEVERITY_SUCCESS,FACILITY_ITF,0x6003);
const HRESULT RIO_E_CREATE_BITMAP_FAILED     = -1*MAKE_HRESULT(SEVERITY_SUCCESS,FACILITY_ITF,0x6004);
const HRESULT RIO_E_MAX_FOLDERS              = -1*MAKE_HRESULT(SEVERITY_SUCCESS,FACILITY_ITF,0x6005);
const HRESULT RIO_E_NEW_FOLDER_FAILED        = -1*MAKE_HRESULT(SEVERITY_SUCCESS,FACILITY_ITF,0x6006);
const HRESULT RIO_E_FOLDER_ALLOC_FAILED      = -1*MAKE_HRESULT(SEVERITY_SUCCESS,FACILITY_ITF,0x6007);
const HRESULT RIO_E_FOLDER_WRITE_FAILED      = -1*MAKE_HRESULT(SEVERITY_SUCCESS,FACILITY_ITF,0x6008);
const HRESULT RIO_E_COULDNT_SEND_FOLDER_LOC  = -1*MAKE_HRESULT(SEVERITY_SUCCESS,FACILITY_ITF,0x6009);
const HRESULT RIO_E_FILE_NOT_FOUND           = -1*MAKE_HRESULT(SEVERITY_SUCCESS,FACILITY_ITF,0x600A);
const HRESULT RIO_E_NOT_ENOUGH_MEMORY        = -1*MAKE_HRESULT(SEVERITY_SUCCESS,FACILITY_ITF,0x600B);
const HRESULT RIO_E_SHORT_READ               = -1*MAKE_HRESULT(SEVERITY_SUCCESS,FACILITY_ITF,0x600C);
const HRESULT RIO_E_WRITE_FAILED             = -1*MAKE_HRESULT(SEVERITY_SUCCESS,FACILITY_ITF,0x600D);
const HRESULT RIO_E_COULDNT_ACCEPT_DATA      = -1*MAKE_HRESULT(SEVERITY_SUCCESS,FACILITY_ITF,0x600E);
const HRESULT RIO_E_STRING_ERROR             = -1*MAKE_HRESULT(SEVERITY_SUCCESS,FACILITY_ITF,0x600F);
const HRESULT RIO_E_BITMAP_INVALID           = -1*MAKE_HRESULT(SEVERITY_SUCCESS,FACILITY_ITF,0x6010);
const HRESULT RIO_E_NOT_ENOUGH_BITMAP_DATA   = -1*MAKE_HRESULT(SEVERITY_SUCCESS,FACILITY_ITF,0x6011);
const HRESULT RIO_E_INVALID_FOLDER           = -1*MAKE_HRESULT(SEVERITY_SUCCESS,FACILITY_ITF,0x6012);
const HRESULT RIO_E_INVALID_SONG             = -1*MAKE_HRESULT(SEVERITY_SUCCESS,FACILITY_ITF,0x6013);
const HRESULT RIO_E_USER_ABORTED             = -1*MAKE_HRESULT(SEVERITY_SUCCESS,FACILITY_ITF,0x6014);

// Some constants we use again and again...
const unsigned RIO_TEXTGRAPHICSHEIGHT = 16;
const unsigned RIO_TEXTGRAPHICSWIDTH = 768;
const unsigned RIO_BITMAPSIZE = 1536;
const unsigned RIO_BLOCKSIZE = 0x4000;
const unsigned RIO_BIGBLOCKSIZE = 0x10000;
const unsigned RIO_HUGEBLOCKSIZE = 0x100000;

// Our callback function
typedef bool (*ProgressCallback) (int percent, void *param);

struct RioBitmapData
{
  WORD    wNumBlocks;
  BYTE    Image[RIO_BITMAPSIZE];
};

struct RioSongEntry
{
  WORD            wOffset;
  WORD            wDunno1;
  DWORD           dwLength;
  WORD            wDunno2;
  WORD            wDunno3;
  DWORD           dwMP3Sig;
  DWORD           dwTime;
  RioBitmapData   Bitmap; 
  char            szName1[362];
  char            szName2[128];
};

struct RioFolderEntry
{
  DWORD            wOffset;
  //WORD            wDunno1;
  DWORD            wFirstFreeEntryOfs;
  //WORD            wDunno2;
  DWORD           dwDunno3;
  DWORD           dwDunno4;
  DWORD           dwTime;
  RioBitmapData   Bitmap; 
  char            szName1[362];
  char            szName2[128];
};

struct RioMemStatus
{
  WORD    wDunno1;
  WORD    wBlockSize;
  WORD    wNumBlocks;
  WORD    wFirstFreeBlock;
  WORD    wNumUnusedBlocks;
  DWORD   dwDunno2;
  DWORD   dwDunno3;
};

struct RioFolderLocation
{
  WORD    wOffset;
  WORD    wBytes;
  WORD    wFolderNum; 
};

struct UsbDevice
{
  BYTE bLength;
  BYTE bDescriptorType;
  WORD bcdUSB;
  BYTE bDeviceClass;
  BYTE bDeviceSubClass;
  BYTE bDeviceProtocol;
  BYTE bMaxPacketSize;
  WORD idVendor;
  WORD idProduct;
  WORD bcdDevice;
  BYTE iManufacturer;
  BYTE iProduct;
  BYTE iSerialNumber;
  BYTE bNumConfigurations;
};


typedef std::vector<RioFolderEntry> RioFolderList;
typedef std::vector<RioSongEntry> RioSongList;

enum RIO_DIRECTION
{
  DIR_IN,
  DIR_OUT,
  DIR_STANDARD
};

enum RIO_REQUEST
{
  REQ_GET_STATUS              = 0x00,
  REQ_CLEAR_FEATURE           = 0x01,
  REQ_SET_FEATURE             = 0x03,
  REQ_SET_ADDRESS             = 0x05,
  REQ_GET_DESCRIPTOR          = 0x06,
  REQ_SET_DESCRIPTOR          = 0x07,
  REQ_GET_CONFIGURATION       = 0x08, 
  REQ_SET_CONFIGURATION       = 0x09, 
  REQ_GET_INTERFACE           = 0x10,
  REQ_SET_INTERFACE           = 0x11,
  REQ_SYNCH_FRAME             = 0x12,
  REQ_UPLOAD_ANIMATION        = 0x3f,
  REQ_FIRMWARE				        = 0x40,
  REQ_NOP                     = 0x42,
  REQ_QUERY_OFFSET_LAST_WRITE = 0x43,
  REQ_TEST                    = 0x44,
  REQ_READ_FROM_USB           = 0x45,
  REQ_WRITE_TO_USB            = 0x46,
  REQ_START_USB_COMM          = 0x47,
  REQ_END_USB_COMM            = 0x48,
  REQ_SET_READ_ADDRESS        = 0x4B,
  REQ_SET_WRITE_ADDRESS       = 0x4C,
  REQ_RIO_FORMAT_DEVICE       = 0x4D,
  REQ_QUERY_FOLDER_BLOCK      = 0x4E,
  REQ_MYSTERY                 = 0x4F,
  REQ_QUERY_FREE_MEM          = 0x50,
  REQ_SET_TO_MEM_LOCATION     = 0x51,
  REQ_SEND_FOLDER_LOCATION    = 0x56,
  REQ_GET_MEM_STATUS          = 0x57,
  REQ_END_FOLDER_TRANSFERS    = 0x58,
  REQ_QUERY_NUM_FOLDER_BLOCKS = 0x59
};
  

enum USB_DESCRIPTOR
{
  DESC_DEVICE             = 0x0100,
  DESC_CONFIGURATION      = 0x0200,
  DESC_STRING             = 0x0300,
  DESC_INTERFACE          = 0x0400,
  DESC_ENDPOINT           = 0x0500,
  DESC_DEVICE_QUALIFIER   = 0x0600,
  DESC_OTHER_SPEED_CONFIG = 0x0700,
  DESC_INTERFACE_POWER    = 0x0800
};



enum RIO_STATUS
{
  RSTATUS_ERROR = 0,
  RSTATUS_OK = 1
};

class Rio500
{
    public:
  Rio500();
  ~Rio500();

  bool connect();
  bool disconnect();

  bool hasExternalCard();

  int getAvailableMemory(int card);
  int getTotalMemory(int card);
  int getFirmwareVersion();
  bool format(int card);

  bool addFolder(char *name, int card);
  bool renameFolder(int folderNum, char *newName, int card);
  bool deleteFolder(int folderNum, int card);
  RioFolderList getFolders(int card);
  bool putFolders(RioFolderList list, int card);

  bool addSong(int folder, char *name, char *filename, int card, ProgressCallback callback, void *parm);
  bool renameSong(int folder, int song, char *newname, int card);
  bool deleteSong(int folder, int song, int card);
  RioSongList getSongs(int folder, int card);
  bool putSongs(int folder, RioSongList list, int card);

  void getFontName(char *name);
  void setFontName(char *name);
  int  getFontSize();
  void setFontSize(int size);

  bool getSong(int folder, int song, char *filename, int flash, ProgressCallback callback, void *parm);

  HRESULT getLastError();

    protected:

  inline bool IsDriverOpened()
      { return rio500 != INVALID_HANDLE_VALUE; }

  // In the order they were written... all of the protected methods of this object...

  int openDriver();
  int openDriverRaw();
  void closeDriver();

  bool startComm();
  bool endComm();

  DWORD sendCommand(RIO_REQUEST Req, WORD wValue, WORD wIndex);

  bool IORequest(RIO_DIRECTION Dir, RIO_REQUEST Req, WORD wValue, WORD wIndex, DWORD dwLen, 
      void* pData);

  bool buildTextBitmap(char *bsText, BYTE* pBitmapData, WORD& wBlocks);

  bool isFirstFolder(long lFlash);
  RioFolderList readFolderEntries(long lFlash);

  bool allocateEmptyFolderBlock(DWORD dwNumFolders, long lFlash);
  
  void clearBlock(BYTE *pBlock);

  bool  sendWriteCommand(DWORD dwAddress, DWORD dwNumBlocks, long lFlash);
  DWORD bulkWrite(void* pBlock, DWORD dwNumBytes);

  bool writeFolderEntries(RioFolderList& Folders, long lFlash);
  bool sendFolderLocation(DWORD dwOffset, DWORD dwFolderNum, long lFlash);

  bool  sendReadCommand(DWORD dwAddress, DWORD dwNumBlocks, long lFlash);
  DWORD bulkRead(void* pBlock, DWORD dwNumBytes);

  RioSongList readSongEntries(RioFolderEntry& Folder, DWORD dwFolderNum, long lFlash);
  bool writeSongEntries(DWORD dwFolderNum, RioSongList& Songs, long lFlash);

  inline void NOP() { sendCommand(REQ_NOP, 0x0000, 0x00); }

  HANDLE rio500;
  HANDLE rio500Read;
  HANDLE rio500Write;
  char fontName[256];
  int  fontSize;

  UsbDevice deviceInfo;

  HRESULT lastError;
};